@@ -1,4 +1,4 @@
-/* Copyright 1995-2004,2006-2015,2017-2020
+/* Copyright 1995-2004,2006-2015,2017-2020,2022
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -28,10 +28,7 @@
 
 #include <stdio.h>
 #include <fcntl.h>
-
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif
 #include <unistd.h>
 #ifdef HAVE_IO_H
 #include <io.h>
@@ -62,6 +59,9 @@
 #include "pairs.h"
 #include "ports-internal.h"
 #include "posix.h"
+#ifdef __MINGW32__
+# include "posix-w32.h"
+#endif
 #include "read.h"
 #include "strings.h"
 #include "symbols.h"
@@ -208,6 +208,9 @@ scm_i_mode_to_open_flags (SCM mode, int *is_binary, const char *FUNC_NAME)
 	  flags |= O_BINARY;
 #endif
 	  break;
+        case 'e':
+          flags |= O_CLOEXEC;
+          break;
 	case '0':  /* unbuffered: handled later.  */
 	case 'l':  /* line buffered: handled during output.  */
 	  break;
@@ -368,6 +371,9 @@ SCM_DEFINE (scm_i_open_file, "open-file", 2, 0, 1,
 	    "@item +\n"
 	    "Open the port for both input and output.  E.g., @code{r+}: open\n"
 	    "an existing file for both input and output.\n"
+            "@item e\n"
+            "Mark the underlying file descriptor as close-on-exec, as per the\n"
+            "@code{O_CLOEXEC} flag.\n"
 	    "@item 0\n"
 	    "Create an \"unbuffered\" port.  In this case input and output\n"
 	    "operations are passed directly to the underlying port\n"
@@ -480,7 +486,17 @@ fport_input_waiting (SCM port)
   if (poll (&pollfd, 1, 0) < 0)
     scm_syserror ("fport_input_waiting");
 
-  return pollfd.revents & POLLIN ? 1 : 0;
+  if ((pollfd.revents & POLLIN) == 0)
+    return 0;
+
+#ifdef __MINGW32__
+  /* Work around Windows 11 bug where there's always a return character
+   * in the console input queue. */
+  if (console_has_return_keyevent_w32 (fdes))
+    return 0;
+#endif
+
+  return 1;
 }
 
 
@@ -551,6 +567,7 @@ SCM_DEFINE (scm_adjust_port_revealed_x, "adjust-port-revealed!", 2, 0, 0,
 
 
 
+#define FUNC_NAME "fport_print"
 static int 
 fport_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
 {
@@ -567,12 +584,27 @@ fport_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
       scm_putc (' ', port);
       fdes = (SCM_FSTREAM (exp))->fdes;
 
-#if (defined HAVE_TTYNAME) && (defined HAVE_POSIX)
-      if (isatty (fdes))
-	scm_display (scm_ttyname (exp), port);
+#if (!defined HAVE_TTYNAME) || (!defined HAVE_POSIX)
+      scm_intprint (fdes, 10, port);
+#else
+      if (!isatty (fdes))
+        scm_intprint (fdes, 10, port);
       else
-#endif /* HAVE_TTYNAME */
-	scm_intprint (fdes, 10, port);
+        {
+          char *name = 0;
+          SCM_I_LOCKED_SYSCALL(&scm_i_misc_mutex,
+                               char *n = ttyname (fdes);
+                               if (n) name = strdup (n));
+          if (name)
+            scm_display (scm_take_locale_string (name), port);
+          else if (errno == ENODEV)
+            // In some situations ttyname may return ENODEV even though
+            // isatty is true.  See GNU/Linux ttyname(3) as an example.
+            scm_intprint (fdes, 10, port);
+          else
+            SCM_SYSERROR;
+        }
+#endif // (defined HAVE_TTYNAME) && (defined HAVE_POSIX)
     }
   else
     {
@@ -583,6 +615,7 @@ fport_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
   scm_putc ('>', port);
   return 1;
 }
+#undef FUNC_NAME
 
 /* fill a port's read-buffer with a single read.  returns the first
    char or EOF if end of file.  */
