@@ -1,4 +1,4 @@
-/* Copyright 1995-2004,2006,2009-2014,2016-2019
+/* Copyright 1995-2004,2006,2009-2014,2016-2021,2023
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -31,7 +31,13 @@
 #include <sys/stat.h>
 #include <unistd.h>
 
+#include "scm.h"
+
+#if SCM_ENABLE_MINI_GMP
+#include "mini-gmp.h"
+#else
 #include <gmp.h>
+#endif
 
 /* Everybody has an init function.  */
 #include "alist.h"
@@ -46,6 +52,7 @@
 #include "chars.h"
 #include "continuations.h"
 #include "control.h"
+#include "custom-ports.h"
 #include "debug.h"
 #ifdef GUILE_DEBUG_MALLOC
 #include "debug-malloc.h"
@@ -58,6 +65,7 @@
 #include "error.h"
 #include "eval.h"
 #include "evalext.h"
+#include "exceptions.h"
 #include "expand.h"
 #include "extensions.h"
 #include "fdes-finalizers.h"
@@ -70,7 +78,6 @@
 #include "fports.h"
 #include "frames.h"
 #include "gc.h"
-#include "generalized-arrays.h"
 #include "generalized-vectors.h"
 #include "gettext.h"
 #include "goops.h"
@@ -121,7 +128,6 @@
 #include "socket.h"
 #include "sort.h"
 #include "srcprop.h"
-#include "srfi-1.h"
 #include "srfi-13.h"
 #include "srfi-14.h"
 #include "srfi-4.h"
@@ -136,7 +142,6 @@
 #include "symbols.h"
 #include "syntax.h"
 #include "throw.h"
-#include "trees.h"
 #include "unicode.h"
 #include "uniform.h"
 #include "values.h"
@@ -144,7 +149,6 @@
 #include "vectors.h"
 #include "version.h"
 #include "vm.h"
-#include "vports.h"
 #include "weak-set.h"
 #include "weak-table.h"
 #include "weak-vector.h"
@@ -155,49 +159,19 @@
 
 /* initializing standard and current I/O ports */
 
-typedef struct
-{
-  int fdes;
-  char *mode;
-} stream_body_data;
-
-/* proc to be called in scope of exception handler stream_handler. */
-static SCM
-stream_body (void *data)
-{
-  stream_body_data *body_data = (stream_body_data *) data;
-  SCM port = scm_fdes_to_port (body_data->fdes, body_data->mode, SCM_BOOL_F);
-  scm_set_port_revealed_x (port, SCM_INUM1);
-  return port;
-}
-
-/* exception handler for stream_body.  */
-static SCM
-stream_handler (void *data SCM_UNUSED,
-		SCM tag SCM_UNUSED,
-		SCM throw_args SCM_UNUSED)
-{
-  return SCM_BOOL_F;
-}
-
 /* Convert a file descriptor to a port, using scm_fdes_to_port.
    - set the revealed count for FILE's file descriptor to 1, so
-   that fdes won't be closed when the port object is GC'd.
-   - catch exceptions: allow Guile to be able to start up even
-   if it has been handed bogus stdin/stdout/stderr.  replace the
-   bad ports with void ports.  */
+   that fdes won't be closed when the port object is GC'd.  */
 static SCM
 scm_standard_stream_to_port (int fdes, char *mode)
 {
-  SCM port;
-  stream_body_data body_data;
+  long mode_bits = scm_mode_bits (mode);
 
-  body_data.fdes = fdes;
-  body_data.mode = mode;
-  port = scm_internal_catch (SCM_BOOL_T, stream_body, &body_data, 
-			     stream_handler, NULL);
-  if (scm_is_false (port))
-    port = scm_void_port (mode);
+  if (!scm_i_fdes_is_valid (fdes, mode_bits))
+    return scm_void_port (mode);
+
+  SCM port = scm_i_fdes_to_port (fdes, mode_bits, SCM_BOOL_F, 0);
+  scm_set_port_revealed_x (port, SCM_INUM1);
   return port;
 }
 
@@ -398,10 +372,10 @@ scm_i_init_guile (void *base)
   scm_bootstrap_programs ();
   scm_bootstrap_vm ();
   scm_register_atomic ();
+  scm_register_custom_ports ();
   scm_register_fdes_finalizers ();
   scm_register_foreign ();
   scm_register_foreign_object ();
-  scm_register_srfi_1 ();
   scm_register_srfi_60 ();
   scm_register_poll ();
 
@@ -470,12 +444,11 @@ scm_i_init_guile (void *base)
   scm_init_srcprop ();     /* requires smob_prehistory */
   scm_init_stackchk ();
 
-  scm_init_generalized_arrays ();
   scm_init_generalized_vectors ();
   scm_init_vectors ();  /* Requires array-handle, */
   scm_init_uniform ();
   scm_init_bitvectors ();  /* Requires smob_prehistory, array-handle */
-  scm_init_srfi_4 ();  /* Requires smob_prehistory, array-handle */
+  scm_bootstrap_srfi_4 ();  /* Requires smob_prehistory, array-handle */
   scm_init_arrays ();    /* Requires smob_prehistory, array-handle */
   scm_init_array_map ();
 
@@ -489,14 +462,13 @@ scm_i_init_guile (void *base)
   scm_init_strorder ();
   scm_init_srfi_13 ();
   scm_init_srfi_14 ();  /* Requires smob_prehistory */
+  scm_init_exceptions ();
   scm_init_throw ();    /* Requires smob_prehistory */
-  scm_init_trees ();
   scm_init_version ();
   scm_init_weak_set ();
   scm_init_weak_table ();
   scm_init_weak_vectors ();
   scm_init_guardians (); /* requires smob_prehistory */
-  scm_init_vports ();
   scm_init_standard_ports ();  /* Requires fports */
   scm_init_expand ();   /* Requires structs */
   scm_init_memoize ();  /* Requires smob_prehistory */
