@@ -1,21 +1,21 @@
-/* Copyright (C) 1995, 1996, 1998, 1999, 2000, 2001, 2004, 2006, 2008,
- *   2009, 2010, 2011, 2012, 2013, 2014 Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1995-1996,1998-2001,2004,2006,2008-2019,2022
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
@@ -24,29 +24,9 @@
 #  include <config.h>
 #endif
 
+#include <stat-time.h>
 #include <string.h>
 #include <stdio.h>
-
-#include "libguile/_scm.h"
-#include "libguile/alist.h"
-#include "libguile/chars.h"
-#include "libguile/dynwind.h"
-#include "libguile/eval.h"
-#include "libguile/fports.h"
-#include "libguile/libpath.h"
-#include "libguile/loader.h"
-#include "libguile/modules.h"
-#include "libguile/read.h"
-#include "libguile/srfi-13.h"
-#include "libguile/strings.h"
-#include "libguile/throw.h"
-
-#include "libguile/validate.h"
-#include "libguile/load.h"
-#include "libguile/fluids.h"
-
-#include "libguile/vm.h" /* for load-compiled/vm */
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -55,12 +35,40 @@
 #include <pwd.h>
 #endif /* HAVE_PWD_H */
 
+#include "alist.h"
+#include "backtrace.h"
+#include "boolean.h"
+#include "chars.h"
+#include "dynwind.h"
+#include "eq.h"
+#include "eval.h"
+#include "filesys.h"
+#include "fluids.h"
+#include "fports.h"
+#include "gsubr.h"
+#include "keywords.h"
+#include "libpath.h"
+#include "list.h"
+#include "loader.h"
+#include "modules.h"
+#include "pairs.h"
+#include "procs.h"
+#include "read.h"
+#include "srfi-13.h"
+#include "strings.h"
+#include "strports.h"
+#include "symbols.h"
+#include "throw.h"
+#include "variable.h"
+#include "version.h"
+#include "vm.h" /* for load-compiled/vm */
+
+#include "load.h"
+
 #ifndef R_OK
 #define R_OK 4
 #endif
 
-#include <stat-time.h>
-
 
 /* Loading a file, given an absolute filename.  */
 
@@ -98,7 +106,7 @@ SCM_DEFINE (scm_primitive_load, "primitive-load", 1, 0, 0,
     SCM port;
 
     port = scm_open_file_with_encoding (filename,
-                                        scm_from_latin1_string ("r"),
+                                        scm_from_latin1_string ("re"),
                                         SCM_BOOL_T, /* guess_encoding */
                                         scm_from_latin1_string ("UTF-8"));
 
@@ -146,7 +154,7 @@ SCM_DEFINE (scm_sys_package_data_dir, "%package-data-dir", 0, 0, 0,
 	    "@samp{/usr/local/share/guile}.")
 #define FUNC_NAME s_scm_sys_package_data_dir
 {
-  return scm_from_locale_string (SCM_PKGDATA_DIR);
+  return scm_from_utf8_string (SCM_PKGDATA_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_PKGDATA_DIR */
@@ -158,7 +166,7 @@ SCM_DEFINE (scm_sys_library_dir, "%library-dir", 0,0,0,
 	    "E.g., may return \"/usr/share/guile/1.3.5\".")
 #define FUNC_NAME s_scm_sys_library_dir
 {
-  return scm_from_locale_string (SCM_LIBRARY_DIR);
+  return scm_from_utf8_string (SCM_LIBRARY_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_LIBRARY_DIR */
@@ -184,7 +192,7 @@ SCM_DEFINE (scm_sys_global_site_dir, "%global-site-dir", 0,0,0,
 	    "E.g., may return \"/usr/share/guile/site\".")
 #define FUNC_NAME s_scm_sys_global_site_dir
 {
-  return scm_from_locale_string (SCM_GLOBAL_SITE_DIR);
+  return scm_from_utf8_string (SCM_GLOBAL_SITE_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_GLOBAL_SITE_DIR */
@@ -197,7 +205,7 @@ SCM_DEFINE (scm_sys_site_ccache_dir, "%site-ccache-dir", 0,0,0,
 	    "E.g., may return \"/usr/lib/guile/" SCM_EFFECTIVE_VERSION "/site-ccache\".")
 #define FUNC_NAME s_scm_sys_site_ccache_dir
 {
-  return scm_from_locale_string (SCM_SITE_CCACHE_DIR);
+  return scm_from_utf8_string (SCM_SITE_CCACHE_DIR);
 }
 #undef FUNC_NAME
 #endif /* SCM_SITE_CCACHE_DIR */
@@ -302,6 +310,9 @@ scm_i_mirror_backslashes (char *path)
 
       while (*p)
 	{
+          /* FIXME: When the locale encoding is Shift_JIS, backslash '\'
+             has a multibyte representation, so this code will
+             misbehave.  */
 	  if (*p == '\\')
 	    *p = '/';
 	  p++;
@@ -331,10 +342,10 @@ scm_init_load_path ()
   else if (env)
     path = scm_parse_path (scm_from_locale_string (env), path);
   else
-    path = scm_list_4 (scm_from_locale_string (SCM_LIBRARY_DIR),
-                       scm_from_locale_string (SCM_SITE_DIR),
-                       scm_from_locale_string (SCM_GLOBAL_SITE_DIR),
-                       scm_from_locale_string (SCM_PKGDATA_DIR));
+    path = scm_list_4 (scm_from_utf8_string (SCM_LIBRARY_DIR),
+                       scm_from_utf8_string (SCM_SITE_DIR),
+                       scm_from_utf8_string (SCM_GLOBAL_SITE_DIR),
+                       scm_from_utf8_string (SCM_PKGDATA_DIR));
 
   env = scm_i_mirror_backslashes (getenv ("GUILE_SYSTEM_COMPILED_PATH"));
   if (env && strcmp (env, "") == 0)
@@ -344,8 +355,8 @@ scm_init_load_path ()
     cpath = scm_parse_path (scm_from_locale_string (env), cpath);
   else
     {
-      cpath = scm_list_2 (scm_from_locale_string (SCM_CCACHE_DIR),
-                          scm_from_locale_string (SCM_SITE_CCACHE_DIR));
+      cpath = scm_list_2 (scm_from_utf8_string (SCM_CCACHE_DIR),
+                          scm_from_utf8_string (SCM_SITE_CCACHE_DIR));
     }
 
 #endif /* SCM_LIBRARY_DIR */
@@ -946,7 +957,7 @@ SCM_DEFINE (scm_search_path, "search-path", 2, 0, 1,
 	    {
 	      require_exts = SCM_CADR (rest);
 	      if (SCM_UNLIKELY (!scm_is_null (SCM_CDDR (rest))))
-		scm_wrong_num_args (scm_from_locale_string (FUNC_NAME));
+		scm_wrong_num_args (scm_from_utf8_string (FUNC_NAME));
 	    }
 	}
       else
@@ -1152,7 +1163,7 @@ SCM_DEFINE (scm_primitive_load_path, "primitive-load-path", 0, 0, 1,
             "depending on the optional second argument,\n"
             "@var{exception_on_not_found}.  If it is @code{#f}, @code{#f}\n"
             "will be returned.  If it is a procedure, it will be called\n"
-            "with no arguments.  Otherwise an error is signalled.")
+            "with no arguments.  Otherwise an error is signaled.")
 #define FUNC_NAME s_scm_primitive_load_path
 {
   SCM filename, exception_on_not_found;
@@ -1272,10 +1283,10 @@ scm_init_eval_in_scheme (void)
   int found_stale_eval_go = 0;
 
   eval_scm = search_path (*scm_loc_load_path,
-                          scm_from_locale_string ("ice-9/eval.scm"),
+                          scm_from_utf8_string ("ice-9/eval.scm"),
                           SCM_EOL, SCM_BOOL_F, &stat_source);
   eval_thunk =
-    load_thunk_from_path (scm_from_locale_string ("ice-9/eval.go"),
+    load_thunk_from_path (scm_from_utf8_string ("ice-9/eval.go"),
                           eval_scm, &stat_source, &found_stale_eval_go);
   
   if (scm_is_true (eval_thunk))
@@ -1333,13 +1344,13 @@ scm_init_load ()
   scm_loc_load_path = SCM_VARIABLE_LOC (scm_c_define ("%load-path", SCM_EOL));
   scm_loc_load_extensions
     = SCM_VARIABLE_LOC (scm_c_define ("%load-extensions",
-				      scm_list_2 (scm_from_locale_string (".scm"),
+				      scm_list_2 (scm_from_utf8_string (".scm"),
 						  scm_nullstr)));
   scm_loc_load_compiled_path
     = SCM_VARIABLE_LOC (scm_c_define ("%load-compiled-path", SCM_EOL));
   scm_loc_load_compiled_extensions
     = SCM_VARIABLE_LOC (scm_c_define ("%load-compiled-extensions",
-				      scm_list_1 (scm_from_locale_string (".go"))));
+				      scm_list_1 (scm_from_utf8_string (".go"))));
   scm_loc_load_hook = SCM_VARIABLE_LOC (scm_c_define ("%load-hook", SCM_BOOL_F));
 
   scm_loc_compile_fallback_path
@@ -1360,7 +1371,7 @@ scm_init_load ()
 
   init_build_info ();
 
-#include "libguile/load.x"
+#include "load.x"
 }
 
 void
@@ -1387,9 +1398,3 @@ scm_init_load_should_auto_compile ()
 }
   
   
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
