@@ -1,21 +1,21 @@
-/* Copyright (C) 1999, 2000, 2001, 2002, 2004, 2006, 2007, 2008, 2009,
- *   2010, 2011, 2012, 2014 Free Software Foundation, Inc.
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1999-2002,2004,2006-2012,2014,2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
@@ -39,17 +39,19 @@
 # include <config.h>
 #endif
 
-#include "libguile/_scm.h"
-#include "libguile/eval.h"
-#include "libguile/arrays.h"
-#include "libguile/array-map.h"
-#include "libguile/feature.h"
-#include "libguile/vectors.h"
-#include "libguile/async.h"
-#include "libguile/dynwind.h"
+#include "array-map.h"
+#include "arrays.h"
+#include "async.h"
+#include "boolean.h"
+#include "dynwind.h"
+#include "eval.h"
+#include "feature.h"
+#include "gsubr.h"
+#include "list.h"
+#include "pairs.h"
+#include "vectors.h"
 
-#include "libguile/validate.h"
-#include "libguile/sort.h"
+#include "sort.h"
 
 /* We have two quicksort variants: one for SCM (#t) arrays and one for
    typed arrays.
@@ -60,14 +62,14 @@
 #define VEC_PARAM   SCM * ra,
 #define GET(i)      ra[(i)*inc]
 #define SET(i, val) ra[(i)*inc] = val
-#include "libguile/quicksort.i.c"
+#include "quicksort.i.c"
 
 #define NAME        quicksorta
 #define INC_PARAM
 #define VEC_PARAM   scm_t_array_handle * const ra,
 #define GET(i)      scm_array_handle_ref (ra, scm_array_handle_pos_1 (ra, i))
 #define SET(i, val) scm_array_handle_set (ra, scm_array_handle_pos_1 (ra, i), val)
-#include "libguile/quicksort.i.c"
+#include "quicksort.i.c"
 
 SCM_DEFINE (scm_restricted_vector_sort_x, "restricted-vector-sort!", 4, 0, 0,
             (SCM vec, SCM less, SCM startpos, SCM endpos),
@@ -643,13 +645,7 @@ SCM_DEFINE (scm_sort_list, "sort-list", 2, 0, 0,
 void
 scm_init_sort ()
 {
-#include "libguile/sort.x"
+#include "sort.x"
 
   scm_add_feature ("sort");
 }
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
