@@ -1,5 +1,5 @@
 ;;; Type analysis on CPS
-;;; Copyright (C) 2014-2021, 2023 Free Software Foundation, Inc.
+;;; Copyright (C) 2014-2021,2023-2024 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software: you can redistribute it and/or modify
 ;;; it under the terms of the GNU Lesser General Public License as
@@ -1008,10 +1008,12 @@ minimum, and maximum."
   (define! result &u64 param param))
 
 (define-type-checker (scm->u64/truncate scm)
-  (check-type scm &exact-integer &range-min &range-max))
+  (check-type scm &exact-integer -inf.0 +inf.0))
 (define-type-inferrer (scm->u64/truncate scm result)
-  (restrict! scm &exact-integer &range-min &range-max)
-  (define! result &u64 0 &u64-max))
+  (restrict! scm &exact-integer -inf.0 +inf.0)
+  (if (<= 0 (&min scm) (&max scm) &u64-max)
+      (define! result &u64 (&min scm) (&max scm))
+      (define! result &u64 0 &u64-max)))
 
 (define-type-checker (u64->scm u64)
   #t)
@@ -1653,11 +1655,8 @@ where (A0 <= A <= A1) and (B0 <= B <= B1)."
       (define-exact-integer! result min max))))
 
 (define-type-inferrer (ulogand a b result)
-  (restrict! a &u64 0 &u64-max)
-  (restrict! b &u64 0 &u64-max)
   (define! result &u64 0 (min (&max/u64 a) (&max/u64 b))))
 (define-type-inferrer/param (ulogand/immediate param a result)
-  (restrict! a &u64 0 &u64-max)
   (call-with-values (lambda ()
                       (logand-bounds (&min a) (&max a) param param))
     (lambda (min max)
@@ -1682,8 +1681,6 @@ i.e. (logand A (lognot B)), where (A0 <= A <= A1) and (B0 <= B <= B1)."
       (define-exact-integer! result min max))))
 
 (define-type-inferrer (ulogsub a b result)
-  (restrict! a &u64 0 &u64-max)
-  (restrict! b &u64 0 &u64-max)
   (define! result &u64 0 (&max/u64 a)))
 
 (define (logior-bounds a0 a1 b0 b1)
@@ -1729,8 +1726,6 @@ where (A0 <= A <= A1) and (B0 <= B <= B1)."
       (define-exact-integer! result min max))))
 
 (define-type-inferrer (ulogior a b result)
-  (restrict! a &u64 0 &u64-max)
-  (restrict! b &u64 0 &u64-max)
   (define! result &u64
     (max (&min/0 a) (&min/0 b))
     (saturate+ (&max/u64 a) (&max/u64 b))))
@@ -1786,8 +1781,6 @@ where (A0 <= A <= A1) and (B0 <= B <= B1)."
       (define! result &exact-integer min max))))
 
 (define-type-inferrer (ulogxor a b result)
-  (restrict! a &u64 0 &u64-max)
-  (restrict! b &u64 0 &u64-max)
   (define! result &u64 0 (saturate+ (&max/u64 a) (&max/u64 b))))
 
 (define-simple-type-checker (lognot &exact-integer))
