@@ -1,4 +1,4 @@
-;;;; 	Copyright (C) 2009-2014, 2017-2018 Free Software Foundation, Inc.
+;;;; 	Copyright (C) 2009-2014,2017-2020,2022-2023 Free Software Foundation, Inc.
 ;;;;
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -21,8 +21,8 @@
   #:use-module (srfi srfi-11)
   #:use-module (ice-9 match)
   #:use-module (system base syntax)
-  #:export (tree-il-src
-
+  #:export ((tree-il-src/ensure-alist . tree-il-src)
+            (tree-il-src . tree-il-srcv)
             <void> void? make-void void-src
             <const> const? make-const const-src const-exp
             <primitive-ref> primitive-ref? make-primitive-ref primitive-ref-src primitive-ref-name
@@ -30,9 +30,9 @@
             <lexical-set> lexical-set? make-lexical-set lexical-set-src lexical-set-name lexical-set-gensym lexical-set-exp
             <module-ref> module-ref? make-module-ref module-ref-src module-ref-mod module-ref-name module-ref-public?
             <module-set> module-set? make-module-set module-set-src module-set-mod module-set-name module-set-public? module-set-exp
-            <toplevel-ref> toplevel-ref? make-toplevel-ref toplevel-ref-src toplevel-ref-name
-            <toplevel-set> toplevel-set? make-toplevel-set toplevel-set-src toplevel-set-name toplevel-set-exp
-            <toplevel-define> toplevel-define? make-toplevel-define toplevel-define-src toplevel-define-name toplevel-define-exp
+            <toplevel-ref> toplevel-ref? make-toplevel-ref toplevel-ref-src toplevel-ref-mod toplevel-ref-name
+            <toplevel-set> toplevel-set? make-toplevel-set toplevel-set-src toplevel-set-mod toplevel-set-name toplevel-set-exp
+            <toplevel-define> toplevel-define? make-toplevel-define toplevel-define-src toplevel-define-mod toplevel-define-name toplevel-define-exp
             <conditional> conditional? make-conditional conditional-src conditional-test conditional-consequent conditional-alternate
             <call> call? make-call call-src call-proc call-args
             <primcall> primcall? make-primcall primcall-src primcall-name primcall-args
@@ -60,6 +60,7 @@
             make-tree-il-folder
             post-order
             pre-order
+            with-lexicals
 
             tree-il=?
             tree-il-hash))
@@ -117,9 +118,9 @@
   ;; (<lexical-set> name gensym exp)
   ;; (<module-ref> mod name public?)
   ;; (<module-set> mod name public? exp)
-  ;; (<toplevel-ref> name)
-  ;; (<toplevel-set> name exp)
-  ;; (<toplevel-define> name exp)
+  ;; (<toplevel-ref> mod name)
+  ;; (<toplevel-set> mod name exp)
+  ;; (<toplevel-define> mod name exp)
   ;; (<conditional> test consequent alternate)
   ;; (<call> proc args)
   ;; (<primcall> name args)
@@ -135,6 +136,20 @@
   (<prompt> escape-only? tag body handler)
   (<abort> tag args tail))
 
+(define tree-il-src/ensure-alist
+  (make-procedure-with-setter
+   (lambda (tree)
+     "Return the source location of TREE as a source property alist."
+     ;; psyntax gives us "source vectors"; convert them lazily to reduce
+     ;; allocations.
+    (match (tree-il-src tree)
+      (#(file line column)
+       `((filename . ,file) (line . ,line) (column . ,column)))
+      (src
+       src)))
+   (lambda (tree src)
+     (set! (tree-il-src tree) src))))
+
 
 
 ;; A helper.
@@ -197,13 +212,13 @@
       (make-module-set loc mod name #f (retrans exp)))
 
      (('toplevel (and name (? symbol?)))
-      (make-toplevel-ref loc name))
+      (make-toplevel-ref loc #f name))
 
      (('set! ('toplevel (and name (? symbol?))) exp)
-      (make-toplevel-set loc name (retrans exp)))
+      (make-toplevel-set loc #f name (retrans exp)))
 
      (('define (and name (? symbol?)) exp)
-      (make-toplevel-define loc name (retrans exp)))
+      (make-toplevel-define loc #f name (retrans exp)))
 
      (('lambda meta body)
       (make-lambda loc meta (retrans body)))
@@ -286,13 +301,13 @@
     (($ <module-set> src mod name public? exp)
      `(set! (,(if public? '@ '@@) ,mod ,name) ,(unparse-tree-il exp)))
 
-    (($ <toplevel-ref> src name)
+    (($ <toplevel-ref> src mod name)
      `(toplevel ,name))
 
-    (($ <toplevel-set> src name exp)
+    (($ <toplevel-set> src mod name exp)
      `(set! (toplevel ,name) ,(unparse-tree-il exp)))
 
-    (($ <toplevel-define> src name exp)
+    (($ <toplevel-define> src mod name exp)
      `(define ,name ,(unparse-tree-il exp)))
 
     (($ <lambda> src meta body)
@@ -356,9 +371,9 @@
                (foldts exp seed ...))
               (($ <module-set> src mod name public? exp)
                (foldts exp seed ...))
-              (($ <toplevel-set> src name exp)
+              (($ <toplevel-set> src mod name exp)
                (foldts exp seed ...))
-              (($ <toplevel-define> src name exp)
+              (($ <toplevel-define> src mod name exp)
                (foldts exp seed ...))
               (($ <conditional> src test consequent alternate)
                (let*-values (((seed ...) (foldts test seed ...))
@@ -449,17 +464,17 @@ This is an implementation of `foldts' as described by Andy Wingo in
                 x
                 (make-module-set src mod name public? exp*))))
 
-         (($ <toplevel-set> src name exp)
+         (($ <toplevel-set> src mod name exp)
           (let ((exp* (lp exp)))
             (if (eq? exp exp*)
                 x
-                (make-toplevel-set src name exp*))))
+                (make-toplevel-set src mod name exp*))))
 
-         (($ <toplevel-define> src name exp)
+         (($ <toplevel-define> src mod name exp)
           (let ((exp* (lp exp)))
             (if (eq? exp exp*)
                 x
-                (make-toplevel-define src name exp*))))
+                (make-toplevel-define src mod name exp*))))
 
          (($ <conditional> src test consequent alternate)
           (let ((test* (lp test))
@@ -568,6 +583,20 @@ This is an implementation of `foldts' as described by Andy Wingo in
 (define (pre-order f x)
   (pre-post-order f (lambda (x) x) x))
 
+(define-syntax-rule (with-lexical src id . body)
+  (let ((k (lambda (id) . body)))
+    (match id
+      (($ <lexical-ref>) (k id))
+      (_
+       (let ((tmp (gensym "v ")))
+         (make-let src (list 'id) (list tmp) (list id)
+                   (k (make-lexical-ref src 'id tmp))))))))
+(define-syntax with-lexicals
+  (syntax-rules ()
+    ((with-lexicals src () . body) (let () . body))
+    ((with-lexicals src (id . ids) . body)
+     (with-lexical src id (with-lexicals src ids . body)))))
+
 ;; FIXME: We should have a better primitive than this.
 (define (struct-nfields x)
   (/ (string-length (symbol->string (struct-layout x))) 2))
